"""
Apps module for small-small-hr
"""
from django.apps import AppConfig


class SmallSmallHrConfig(AppConfig):
    """
    Apps config class
    """
    name = 'small_small_hr'
    app_label = 'small_small_hr'

    def ready(self):
        # pylint: disable=unused-import
        import small_small_hr.signals  # noqa

        # set up app settings
        from django.conf import settings
        import small_small_hr.settings as defaults
        for name in dir(defaults):
            if name.isupper() and not hasattr(settings, name):
                setattr(settings, name, getattr(defaults, name))



"""
Serializers for users app
"""
from django.contrib.auth.models import User

from rest_framework import serializers

from small_small_hr.models import StaffProfile


# pylint: disable=too-many-ancestors
class UserSerializer(serializers.ModelSerializer):
    """
    UserSerializer class
    """

    class Meta:  # pylint:  disable=too-few-public-methods
        """
        meta options
        """
        model = User
        fields = ('username', 'first_name', 'last_name', 'email')


class StaffProfileSerializer(serializers.ModelSerializer):
    """
    Serializer class for StaffProfile model
    """
    first_name = serializers.CharField(source='user.first_name')
    last_name = serializers.CharField(source='user.last_name')
    id_number = serializers.SerializerMethodField()
    phone = serializers.SerializerMethodField()
    sex = serializers.SerializerMethodField()
    role = serializers.SerializerMethodField()
    nhif = serializers.SerializerMethodField()
    nssf = serializers.SerializerMethodField()
    pin_number = serializers.SerializerMethodField()
    address = serializers.SerializerMethodField()
    birthday = serializers.SerializerMethodField()
    leave_days = serializers.SerializerMethodField()
    sick_days = serializers.SerializerMethodField()
    overtime_allowed = serializers.SerializerMethodField()
    start_date = serializers.SerializerMethodField()
    end_date = serializers.SerializerMethodField()
    emergency_contact_name = serializers.SerializerMethodField()
    emergency_contact_number = serializers.SerializerMethodField()

    class Meta:  # pylint:  disable=too-few-public-methods
        """
        class meta options
        """
        model = StaffProfile
        fields = [
            'id',
            'first_name',
            'last_name',
            'created',
            'id_number',
            'phone',
            'sex',
            'modified',
            'role',
            'nhif',
            'emergency_contact_name',
            'nssf',
            'address',
            'birthday',
            'overtime_allowed',
            'leave_days',
            'start_date',
            'sick_days',
            'pin_number',
            'end_date',
            'emergency_contact_number',
        ]

    def get_id_number(self, obj):  # pylint: disable=no-self-use
        """
        Get id_number
        """
        return obj.data.get('id_number')

    def get_phone(self, obj):  # pylint: disable=no-self-use
        """
        Get phone
        """
        return obj.data.get('phone')

    def get_sex(self, obj):  # pylint: disable=no-self-use
        """
        Get sex
        """
        return obj.data.get('sex')

    def get_role(self, obj):  # pylint: disable=no-self-use
        """
        Get role
        """
        return obj.data.get('role')

    def get_nhif(self, obj):  # pylint: disable=no-self-use
        """
        Get nhhf
        """
        return obj.data.get('nhif')

    def get_nssf(self, obj):  # pylint: disable=no-self-use
        """
        Get nssf
        """
        return obj.data.get('nssf')

    def get_pin_number(self, obj):  # pylint: disable=no-self-use
        """
        Get pin_number
        """
        return obj.data.get('pin_number')

    def get_address(self, obj):  # pylint: disable=no-self-use
        """
        Get address
        """
        return obj.data.get('address')

    def get_birthday(self, obj):  # pylint: disable=no-self-use
        """
        Get birthday
        """
        return obj.data.get('birthday')

    def get_leave_days(self, obj):  # pylint: disable=no-self-use
        """
        Get leave_days
        """
        return obj.data.get('leave_days')

    def get_sick_days(self, obj):  # pylint: disable=no-self-use
        """
        Get sick_days
        """
        return obj.data.get('sick_days')

    def get_overtime_allowed(self, obj):  # pylint: disable=no-self-use
        """
        Get overtime_allowed
        """
        return obj.data.get('overtime_allowed')

    def get_start_date(self, obj):  # pylint: disable=no-self-use
        """
        Get start_date
        """
        return obj.data.get('start_date')

    def get_end_date(self, obj):  # pylint: disable=no-self-use
        """
        Get end_date
        """
        return obj.data.get('end_date')

    def get_emergency_contact_name(self, obj):  # pylint: disable=no-self-use
        """
        Get emergency_contact_name
        """
        return obj.data.get('emergency_contact_name')

    def get_emergency_contact_number(self, obj):  # pylint: disable=no-self-use
        """
        Get emergency_contact_number
        """
        return obj.data.get('emergency_contact_number')



"""
Emails module for scam app
"""
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils.translation import ugettext as _

from small_small_hr.models import Leave, OverTime


def send_email(  # pylint: disable=too-many-arguments,too-many-locals,bad-continuation
    name: str,
    email: str,
    subject: str,
    message: str,
    obj: object = None,
    cc_list: list = None,
    template: str = "generic",
    template_path: str = "small_small_hr/email",
):
    """
    Sends a generic email

    :param name: name of person
    :param email: email address to send to
    :param subject: the email's subject
    :param message: the email's body text
    :param obj: the object in question
    :param cc_list: the list of email address to "CC"
    :param template: the template to use
    """
    context = {
        "name": name,
        "subject": subject,
        "message": message,
        "object": obj,
        "SITE": Site.objects.get_current(),
    }
    email_subject = render_to_string(
        f"{template_path}/{template}_email_subject.txt", context
    ).replace("\n", "")
    email_txt_body = render_to_string(
        f"{template_path}/{template}_email_body.txt", context
    )
    email_html_body = render_to_string(
        f"{template_path}/{template}_email_body.html", context
    ).replace("\n", "")

    subject = email_subject
    from_email = settings.DEFAULT_FROM_EMAIL
    to_email = f"{name} <{email}>"
    text_content = email_txt_body
    html_content = email_html_body
    msg = EmailMultiAlternatives(subject, text_content, from_email, [to_email])
    if cc_list:
        msg.cc = cc_list
    msg.attach_alternative(html_content, "text/html")

    return msg.send(fail_silently=True)


def leave_application_email(leave_obj: Leave):
    """
    Sends an email to admins when a leave application is made
    """
    msg = getattr(
        settings,
        "SSHR_LEAVE_APPLICATION_EMAIL_TXT",
        _("There has been a new leave application.  Please log in to process " "it."),
    )
    subj = getattr(
        settings, "SSHR_LEAVE_APPLICATION_EMAIL_SUBJ", _("New Leave Application")
    )
    admin_emails = settings.SSHR_ADMIN_LEAVE_EMAILS

    for admin_email in admin_emails:
        send_email(
            name=settings.SSHR_ADMIN_NAME,
            email=admin_email,
            subject=subj,
            message=msg,
            obj=leave_obj,
            template="leave_application",
        )


def leave_processed_email(leave_obj: Leave):
    """
    Sends an email to admins when a leave application is processed
    """
    if leave_obj.staff.user.email:
        msg = getattr(
            settings,
            "SSHR_LEAVE_PROCESSED_EMAIL_TXT",
            _(
                f"You leave application status is "
                f"{leave_obj.get_status_display()}.  Log in for more info."
            ),
        )
        subj = getattr(
            settings,
            "SSHR_LEAVE_PROCESSED_EMAIL_SUBJ",
            _("Your leave application has been processed"),
        )

        send_email(
            name=leave_obj.staff.get_name(),
            email=leave_obj.staff.user.email,
            subject=subj,
            message=msg,
            obj=leave_obj,
            cc_list=settings.SSHR_ADMIN_LEAVE_EMAILS,
        )


def overtime_application_email(overtime_obj: OverTime):
    """
    Sends an email to admins when an overtime application is made
    """
    msg = getattr(
        settings,
        "SSHR_OVERTIME_APPLICATION_EMAIL_TXT",
        _(
            "There has been a new overtime application.  Please log in to "
            "process it."
        ),
    )
    subj = getattr(
        settings, "SSHR_OVERTIME_APPLICATION_EMAIL_SUBJ", _("New Overtime Application")
    )
    admin_emails = settings.SSHR_ADMIN_OVERTIME_EMAILS

    for admin_email in admin_emails:
        send_email(
            name=settings.SSHR_ADMIN_NAME,
            email=admin_email,
            subject=subj,
            message=msg,
            obj=overtime_obj,
            template="overtime_application",
        )


def overtime_processed_email(overtime_obj: OverTime):
    """
    Sends an email to admins when an overtime application is processed
    """
    if overtime_obj.staff.user.email:

        msg = getattr(
            settings,
            "SSHR_OVERTIME_PROCESSED_EMAIL_TXT",
            _(
                f"You overtime application status is "
                f"{overtime_obj.get_status_display()}.  Log in for more info."
            ),
        )
        subj = getattr(
            settings,
            "SSHR_OVERTIME_PROCESSED_EMAIL_SUBJ",
            _("Your overtime application has been processed"),
        )

        send_email(
            name=overtime_obj.staff.get_name(),
            email=overtime_obj.staff.user.email,
            subject=subj,
            message=msg,
            obj=overtime_obj,
            cc_list=settings.SSHR_ADMIN_OVERTIME_EMAILS,
        )



"""
Setup.py for small_small_hr
"""
from os import path

from setuptools import find_packages, setup

# read the contents of your README file
with open(
        path.join(path.abspath(path.dirname(__file__)), "README.md"),
        encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="small-small-hr",
    version=__import__("small_small_hr").__version__,
    description="Minimal human resource management app for Django",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Kelvin Jayanoris",
    author_email="kelvin@jayanoris.com",
    url="https://github.com/moshthepitt/small-small-hr",
    packages=find_packages(exclude=["docs", "tests"]),
    install_requires=[
        "Django >= 2.0.11",
        "voluptuous",
        "psycopg2-binary",
        "sorl-thumbnail",
        "django-private-storage",
        "phonenumberslite",
        "django-phonenumber-field",
        "django-crispy-forms",
        "djangorestframework",
        "sorl-thumbnail",
        "Pillow",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
    ],
)



"""
Forms module for small small hr
"""
from datetime import datetime, time

from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import Q
from django.utils.translation import ugettext as _

import pytz
from crispy_forms.bootstrap import Field, FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit
from phonenumber_field.formfields import PhoneNumberField
from phonenumber_field.phonenumber import PhoneNumber

from small_small_hr.emails import (leave_application_email,
                                   overtime_application_email)
from small_small_hr.models import (TWOPLACES, AnnualLeave, FreeDay, Leave,
                                   OverTime, Role, StaffDocument, StaffProfile)


class AnnualLeaveForm(forms.ModelForm):
    """
    Form used when managing AnnualLeave
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = AnnualLeave
        fields = [
            'staff',
            'year',
            'leave_type',
            'allowed_days',
            'carried_over_days'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        if not self.instance:
            self.fields['year'].initial = datetime.today().year
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'annual-leave-form'
        self.helper.layout = Layout(
            Field('staff',),
            Field('year',),
            Field('leave_type',),
            Field('allowed_days',),
            Field('carried_over_days'),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class RoleForm(forms.ModelForm):
    """
    Form used when managing Role objects
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = Role
        fields = [
            'name',
            'description'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'role-form'
        self.helper.layout = Layout(
            Field('name',),
            Field('description',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class FreeDayForm(forms.ModelForm):
    """
    Form used when managing FreeDay objects
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = FreeDay
        fields = [
            'name',
            'date'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'freeday-form'
        self.helper.layout = Layout(
            Field('name',),
            Field('date',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class OverTimeForm(forms.ModelForm):
    """
    Form used when managing OverTime objects
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = OverTime
        fields = [
            'staff',
            'date',
            'start',
            'end',
            'reason',
            'status',
            'comments'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'overtime-form'
        self.helper.layout = Layout(
            Field('staff',),
            Field('date',),
            Field('start',),
            Field('end',),
            Field('reason',),
            Field('status',),
            Field('comments'),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )

    def clean(self):
        """
        Custom clean method
        """
        cleaned_data = super().clean()
        end = cleaned_data.get('end')
        start = cleaned_data.get('start')
        date = cleaned_data.get('date')
        staff = cleaned_data.get('staff')
        status = cleaned_data.get('status')

        # end must be later than start
        if end <= start:
            self.add_error('end', _("end must be greater than start"))

        # must not overlap within the same date unless being rejected
        # pylint: disable=no-member
        overlap_qs = OverTime.objects.filter(
            date=date, staff=staff, status=OverTime.APPROVED).filter(
                Q(start__gte=start) & Q(end__lte=end))

        if self.instance is not None:
            overlap_qs = overlap_qs.exclude(id=self.instance.id)

        if overlap_qs.exists() and status != OverTime.REJECTED:
            msg = _('you cannot have overlapping overtime hours on the '
                    'same day')
            self.add_error('start', msg)
            self.add_error('end', msg)
            self.add_error('date', msg)


class ApplyOverTimeForm(OverTimeForm):
    """
    Form used when applying for overtime
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = OverTime
        fields = [
            'staff',
            'date',
            'start',
            'end',
            'reason',
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = kwargs.pop('request', None)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields['staff'].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'overtime-application-form'
        self.helper.layout = Layout(
            Field('staff', type="hidden"),
            Field('date',),
            Field('start',),
            Field('end',),
            Field('reason',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )

    def save(self, commit=True):
        """
        Custom save method
        """
        overtime = super().save()
        overtime_application_email(overtime_obj=overtime)
        return overtime


class LeaveForm(forms.ModelForm):
    """
    Form used when managing Leave objects
    """
    start = forms.DateField(label=_('Start Date'), required=True)
    end = forms.DateField(label=_('End Date'), required=True)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = Leave
        fields = [
            'staff',
            'leave_type',
            'start',
            'end',
            'reason',
            'status',
            'comments'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'leave-form'
        self.helper.layout = Layout(
            Field('staff',),
            Field('leave_type',),
            Field('start',),
            Field('end',),
            Field('reason',),
            Field('status',),
            Field('comments'),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )

    def clean_start(self):
        """
        clean start field
        """
        data = self.cleaned_data['start']
        data = datetime.combine(
            date=data,
            time=time(settings.SSHR_DEFAULT_TIME, 0, 0, 0),
            tzinfo=pytz.timezone(settings.TIME_ZONE))
        return data

    def clean_end(self):
        """
        clean end field
        """
        data = self.cleaned_data['end']
        data = datetime.combine(
            date=data,
            time=time(settings.SSHR_DEFAULT_TIME, 0, 0, 0),
            tzinfo=pytz.timezone(settings.TIME_ZONE))
        return data

    def clean(self):
        """
        Custom clean method
        """
        cleaned_data = super().clean()
        leave_type = cleaned_data.get('leave_type')
        staff = cleaned_data.get('staff')
        end = cleaned_data.get('end')
        start = cleaned_data.get('start')
        status = cleaned_data.get('status')

        if all([staff, leave_type, start, end]):
            # end year and start year must be the same
            if end.year != start.year:
                msg = _('start and end must be from the same year')
                self.add_error('start', msg)
                self.add_error('end', msg)

            # end must be later than start
            if end < start:
                self.add_error('end', _("end must be greater than start"))

            if not settings.SSHR_ALLOW_OVERSUBSCRIBE:
                # staff profile must have sufficient sick days
                if leave_type == Leave.SICK:
                    sick_days = staff.get_available_sick_days(year=start.year)
                    if (end - start).days > sick_days:
                        msg = _('Not enough sick days. Available sick days '
                                f'are {sick_days.quantize(TWOPLACES)}')
                        self.add_error('start', msg)
                        self.add_error('end', msg)

                # staff profile must have sufficient leave days
                if leave_type == Leave.REGULAR:
                    leave_days = staff.get_available_leave_days(
                        year=start.year)
                    if (end - start).days > leave_days:
                        msg = _('Not enough leave days. Available leave days '
                                f'are {leave_days.quantize(TWOPLACES)}')
                        self.add_error('start', msg)
                        self.add_error('end', msg)

            # must not overlap unless it is being rejected
            # pylint: disable=no-member
            overlap_qs = Leave.objects.filter(
                staff=staff,
                status=Leave.APPROVED,
                leave_type=leave_type).filter(
                    Q(start__gte=start) & Q(end__lte=end))

            if self.instance is not None:
                overlap_qs = overlap_qs.exclude(id=self.instance.id)

            if overlap_qs.exists() and status != Leave.REJECTED:
                msg = _('you cannot have overlapping leave days')
                self.add_error('start', msg)
                self.add_error('end', msg)


class ApplyLeaveForm(LeaveForm):
    """
    Form used when applying for Leave
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = Leave
        fields = [
            'staff',
            'leave_type',
            'start',
            'end',
            'reason',
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = kwargs.pop('request', None)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields['staff'].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'leave-application-form'
        self.helper.layout = Layout(
            Field('staff', type="hidden"),
            Field('leave_type',),
            Field('start',),
            Field('end',),
            Field('reason',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )

    def save(self, commit=True):
        """
        Custom save method
        """
        leave = super().save()
        leave_application_email(leave_obj=leave)
        return leave


class StaffDocumentForm(forms.ModelForm):
    """
    Form used when managing StaffDocument objects
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = StaffDocument
        fields = [
            'staff',
            'name',
            'description',
            'public',
            'file',
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.file:
            self.fields['file'].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'staffdocument-form'
        self.helper.layout = Layout(
            Field('staff',),
            Field('name',),
            Field('description',),
            Field('file',),
            Field('public',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class UserStaffDocumentForm(forms.ModelForm):
    """
    Form used when managing one's own StaffDocument objects
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = StaffDocument
        fields = [
            'staff',
            'name',
            'description',
            'file',
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields['staff'].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id)
        if self.instance and self.instance.file:
            self.fields['file'].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'staffdocument-form'
        self.helper.layout = Layout(
            Field('staff', type="hidden"),
            Field('name',),
            Field('description',),
            Field('file',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class StaffProfileAdminForm(forms.ModelForm):
    """
    Form used when managing StaffProfile objects
    """
    first_name = forms.CharField(label=_('First Name'), required=True)
    last_name = forms.CharField(label=_('Last Name'), required=True)
    id_number = forms.CharField(label=_('ID Number'), required=True)
    nhif = forms.CharField(label=_('NHIF'), required=False)
    nssf = forms.CharField(label=_('NSSF'), required=False)
    pin_number = forms.CharField(label=_('PIN Number'), required=False)
    emergency_contact_name = forms.CharField(
        label=_('Emergency Contact Name'), required=False)
    emergency_contact_relationship = forms.CharField(
        label=_('Emergency Contact Relationship'), required=False)
    emergency_contact_number = PhoneNumberField(
        label=_('Emergency Contact Phone Number'), required=False)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = StaffProfile
        fields = [
            'first_name',
            'last_name',
            'id_number',
            'image',
            'phone',
            'sex',
            'role',
            'nhif',
            'nssf',
            'pin_number',
            'address',
            'birthday',
            'leave_days',
            'sick_days',
            'overtime_allowed',
            'start_date',
            'end_date',
            'emergency_contact_name',
            'emergency_contact_number',
            'emergency_contact_relationship'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.image:
            self.fields['image'].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'staffprofile-form'
        self.helper.layout = Layout(
            Field('first_name',),
            Field('last_name',),
            Field('image',),
            Field('phone',),
            Field('id_number',),
            Field('sex',),
            Field('role',),
            Field('nhif',),
            Field('nssf',),
            Field('pin_number',),
            Field('address',),
            Field('birthday',),
            Field('leave_days',),
            Field('sick_days',),
            Field('overtime_allowed',),
            Field('start_date',),
            Field('end_date',),
            Field('emergency_contact_name',),
            Field('emergency_contact_number',),
            Field('emergency_contact_relationship',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )

    def clean_id_number(self):
        """
        Check if id number is unique
        """
        value = self.cleaned_data.get('id_number')
        # pylint: disable=no-member
        if StaffProfile.objects.exclude(
                id=self.instance.id).filter(data__id_number=value).exists():
            raise forms.ValidationError(
                _('This id number is already in use.'))
        return value

    def clean_nssf(self):
        """
        Check if NSSF number is unique
        """
        value = self.cleaned_data.get('nssf')
        # pylint: disable=no-member
        if value and StaffProfile.objects.exclude(
                id=self.instance.id).filter(data__nssf=value).exists():
            raise forms.ValidationError(
                _('This NSSF number is already in use.'))
        return value

    def clean_nhif(self):
        """
        Check if NHIF number is unique
        """
        value = self.cleaned_data.get('nhif')
        # pylint: disable=no-member
        if value and StaffProfile.objects.exclude(
                id=self.instance.id).filter(data__nhif=value).exists():
            raise forms.ValidationError(
                _('This NHIF number is already in use.'))
        return value

    def clean_pin_number(self):
        """
        Check if PIN number is unique
        """
        value = self.cleaned_data.get('pin_number')
        # pylint: disable=no-member
        if value and StaffProfile.objects.exclude(
                id=self.instance.id).filter(data__pin_number=value).exists():
            raise forms.ValidationError(
                _('This PIN number is already in use.'))
        return value

    def save(self, commit=True):  # pylint: disable=unused-argument
        """
        Custom save method
        """
        staffprofile = super().save()

        emergency_phone = self.cleaned_data.get('emergency_contact_number')
        if isinstance(emergency_phone, PhoneNumber):
            emergency_phone = emergency_phone.as_e164

        json_data = {
            'id_number': self.cleaned_data.get('id_number'),
            'nhif': self.cleaned_data.get('nhif'),
            'nssf': self.cleaned_data.get('nssf'),
            'pin_number': self.cleaned_data.get('pin_number'),
            'emergency_contact_name': self.cleaned_data.get(
                'emergency_contact_name'),
            'emergency_contact_relationship': self.cleaned_data.get(
                'emergency_contact_relationship'),
            'emergency_contact_number': emergency_phone,
        }
        staffprofile.data = json_data
        staffprofile.save()

        user = staffprofile.user
        user.first_name = self.cleaned_data['first_name']
        user.last_name = self.cleaned_data['last_name']
        user.save()
        return staffprofile


class StaffProfileAdminCreateForm(StaffProfileAdminForm):
    """
    Form used when creating new Staff Profiles
    """
    user = forms.ModelChoiceField(
        label=_('User'), queryset=User.objects.filter(staffprofile=None))

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = StaffProfile
        fields = [
            'user',
            'first_name',
            'last_name',
            'id_number',
            'image',
            'phone',
            'sex',
            'role',
            'nhif',
            'nssf',
            'pin_number',
            'address',
            'birthday',
            'leave_days',
            'sick_days',
            'overtime_allowed',
            'start_date',
            'end_date',
            'emergency_contact_name',
            'emergency_contact_number',
            'emergency_contact_relationship'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.image:
            self.fields['image'].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'staffprofile-form'
        self.helper.layout = Layout(
            Field('user',),
            Field('first_name',),
            Field('last_name',),
            Field('image',),
            Field('phone',),
            Field('id_number',),
            Field('sex',),
            Field('role',),
            Field('nhif',),
            Field('nssf',),
            Field('pin_number',),
            Field('address',),
            Field('birthday',),
            Field('leave_days',),
            Field('sick_days',),
            Field('overtime_allowed',),
            Field('start_date',),
            Field('end_date',),
            Field('emergency_contact_name',),
            Field('emergency_contact_number',),
            Field('emergency_contact_relationship',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )


class StaffProfileUserForm(StaffProfileAdminForm):
    """
    Form used when the user is updating their own data
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Class meta options
        """
        model = StaffProfile
        fields = [
            'first_name',
            'last_name',
            'id_number',
            'image',
            'phone',
            'sex',
            'nhif',
            'nssf',
            'pin_number',
            'address',
            'birthday',
            'emergency_contact_name',
            'emergency_contact_number',
            'emergency_contact_relationship'
        ]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'post'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = 'staffprofile-user-form'
        self.helper.layout = Layout(
            Field('first_name',),
            Field('last_name',),
            Field('image',),
            Field('phone',),
            Field('id_number',),
            Field('sex',),
            Field('nhif',),
            Field('nssf',),
            Field('pin_number',),
            Field('address',),
            Field('birthday',),
            Field('emergency_contact_name',),
            Field('emergency_contact_number',),
            Field('emergency_contact_relationship',),
            FormActions(
                Submit('submitBtn', _('Submit'), css_class='btn-primary'),
            )
        )



"""
Models module for small_small_hr
"""
from datetime import datetime, timedelta
from decimal import Decimal

from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import ugettext as _

from phonenumber_field.modelfields import PhoneNumberField
from private_storage.fields import PrivateFileField
from sorl.thumbnail import ImageField

from small_small_hr.managers import LeaveManager

USER = settings.AUTH_USER_MODEL
TWOPLACES = Decimal(10) ** -2


class TimeStampedModel(models.Model):
    """
    Abstract model class that includes timestamp fields
    """
    created = models.DateTimeField(
        verbose_name=_('Created'),
        auto_now_add=True)
    modified = models.DateTimeField(
        verbose_name=_('Modified'),
        auto_now=True)

    # pylint: disable=too-few-public-methods
    class Meta:
        """
        Meta options for TimeStampedModel
        """
        abstract = True


class Role(TimeStampedModel, models.Model):
    """
    Model class for staff member role
    """
    name = models.CharField(_('Name'), max_length=255)
    description = models.TextField(_('Description'), blank=True, default='')

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for StaffDocument
        """
        abstract = False
        verbose_name = _('Role')
        verbose_name_plural = _('Roles')
        ordering = ['name', 'created']

    def __str__(self):
        # pylint: disable=no-member
        return self.name


class StaffProfile(TimeStampedModel, models.Model):
    """
    StaffProfile model class
    Extends auth.User and adds more fields
    """

    # sex choices
    # according to https://en.wikipedia.org/wiki/ISO/IEC_5218
    NOT_KNOWN = '0'
    MALE = '1'
    FEMALE = '2'
    NOT_APPLICABLE = '9'

    SEX_CHOICES = (
        (NOT_KNOWN, _('Not Known')),
        (MALE, _('Male')),
        (FEMALE, _('Female')),
        (NOT_APPLICABLE, _('Not Applicable'))
    )

    user = models.OneToOneField(
        USER, verbose_name=_('User'), on_delete=models.CASCADE)
    image = ImageField(upload_to="staff-images/", max_length=255,
                       verbose_name=_("Profile Image"),
                       help_text=_("A square image works best"), blank=True)
    sex = models.CharField(_('Gender'), choices=SEX_CHOICES, max_length=1,
                           default=NOT_KNOWN, blank=True, db_index=True)
    role = models.ForeignKey(Role, verbose_name=_('Role'), blank=True,
                             default=None, null=True,
                             on_delete=models.SET_NULL)
    phone = PhoneNumberField(_('Phone'), blank=True, default='')
    address = models.TextField(_('Addresss'), blank=True, default="")
    birthday = models.DateField(_('Birthday'), blank=True, default=None,
                                null=True)
    leave_days = models.PositiveIntegerField(
        _('Leave days'), default=21, blank=True,
        help_text=_('Number of leave days allowed in a year.'))
    sick_days = models.PositiveIntegerField(
        _('Sick days'), default=10, blank=True,
        help_text=_('Number of sick days allowed in a year.'))
    overtime_allowed = models.BooleanField(
        _('Overtime allowed'), blank=True, default=False)
    start_date = models.DateField(
        _('Start Date'), null=True, default=None, blank=True,
        help_text=_('The start date of employment'))
    end_date = models.DateField(
        _('End Date'), null=True, default=None, blank=True,
        help_text=_('The end date of employment'))
    data = JSONField(_('Data'), default=dict, blank=True)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for StaffProfile
        """
        abstract = False
        verbose_name = _('Staff Profile')
        verbose_name_plural = _('Staff Profiles')
        ordering = ['user__first_name', 'user__last_name', 'user__username',
                    'created']

    def get_name(self):
        """
        Returns the staff member's name
        """
        # pylint: disable=no-member
        return f'{self.user.first_name} {self.user.last_name}'

    def get_approved_leave_days(self, year: int = datetime.today().year):
        """
        Get approved leave days in the current year
        """
        # pylint: disable=no-member
        return get_taken_leave_days(
            staffprofile=self,
            status=Leave.APPROVED,
            leave_type=Leave.REGULAR,
            start_year=year,
            end_year=year
        )

    def get_approved_sick_days(self, year: int = datetime.today().year):
        """
        Get approved leave days in the current year
        """
        return get_taken_leave_days(
            staffprofile=self,
            status=Leave.APPROVED,
            leave_type=Leave.SICK,
            start_year=year,
            end_year=year
        )

    def get_available_leave_days(self, year: int = datetime.today().year):
        """
        Get available leave days
        """
        try:
            # pylint: disable=no-member
            leave_record = AnnualLeave.objects.get(
                leave_type=Leave.REGULAR,
                staff=self,
                year=year)
        except AnnualLeave.DoesNotExist:
            return Decimal(0)
        else:
            return leave_record.get_available_leave_days()

    def get_available_sick_days(self, year: int = datetime.today().year):
        """
        Get available sick days
        """
        try:
            # pylint: disable=no-member
            leave_record = AnnualLeave.objects.get(
                leave_type=Leave.SICK,
                staff=self,
                year=year)
        except AnnualLeave.DoesNotExist:
            return Decimal(0)
        else:
            return leave_record.get_available_leave_days()

    def __str__(self):
        return self.get_name()  # pylint: disable=no-member


class StaffDocument(TimeStampedModel, models.Model):
    """
    StaffDocument model class
    """
    staff = models.ForeignKey(
        StaffProfile, verbose_name=_('Staff Member'), on_delete=models.CASCADE)
    name = models.CharField(_('Name'), max_length=255)
    description = models.TextField(_('Description'), blank=True, default='')
    file = PrivateFileField(
        _('File'), upload_to='staff-documents/',
        help_text=_("Upload staff member document"),
        content_types=[
            'application/pdf',
            'application/msword',
            'application/vnd.oasis.opendocument.text',
            'image/jpeg',
            'image/png'
        ],
        max_file_size=10485760
    )
    public = models.BooleanField(
        _('Public'),
        help_text=_('If public, it will be available to everyone.'),
        blank=True, default=False)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for StaffDocument
        """
        abstract = False
        verbose_name = _('Staff Document')
        verbose_name_plural = _('Staff Documents')
        ordering = ['staff', 'name', '-created']

    def __str__(self):
        # pylint: disable=no-member
        return f'{self.staff.get_name()} - {self.name}'


class BaseStaffRequest(TimeStampedModel, models.Model):
    """
    Abstract model class for Leave & Overtime tracking
    """
    APPROVED = '1'
    REJECTED = '2'
    PENDING = '3'

    STATUS_CHOICES = (
        (APPROVED, _('Approved')),
        (PENDING, _('Pending')),
        (REJECTED, _('Rejected'))
    )

    staff = models.ForeignKey(
        StaffProfile, verbose_name=_('Staff Member'), on_delete=models.CASCADE)
    start = models.DateTimeField(_('Start Date'))
    end = models.DateTimeField(_('End Date'))
    reason = models.TextField(_('Reason'), blank=True, default='')
    status = models.CharField(
        _('Status'), max_length=1, choices=STATUS_CHOICES, default=PENDING,
        blank=True, db_index=True)
    comments = models.TextField(_('Comments'), blank=True, default='')

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for StaffDocument
        """
        abstract = True


class Leave(BaseStaffRequest):
    """
    Leave model class
    """
    SICK = '1'
    REGULAR = '2'

    TYPE_CHOICES = (
        (SICK, _('Sick Leave')),
        (REGULAR, _('Regular Leave')),
    )

    leave_type = models.CharField(
        _('Type'), max_length=1, choices=TYPE_CHOICES, default=REGULAR,
        blank=True, db_index=True)

    objects = LeaveManager()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for Leave
        """
        abstract = False
        verbose_name = _('Leave')
        verbose_name_plural = _('Leave')
        ordering = ['staff', '-start']

    def __str__(self):
        # pylint: disable=no-member
        return _(f'{self.staff.get_name()}: {self.start} to {self.end}')


class OverTime(BaseStaffRequest):
    """
    Overtime model class
    """
    date = models.DateField(
        _('Date'), auto_now=False, auto_now_add=False, db_index=True)
    start = models.TimeField(_('Start'), auto_now=False, auto_now_add=False)
    end = models.TimeField(_('End'), auto_now=False, auto_now_add=False)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for OverTime
        """
        abstract = False
        verbose_name = _('Overtime')
        verbose_name_plural = _('Overtime')
        ordering = ['staff', '-date', 'start']

    def __str__(self):
        name = self.staff.get_name()  # pylint: disable=no-member
        return _(f'{name}: {self.date} from {self.start} to {self.end}')

    def get_duration(self):
        """
        Get duration
        """
        start = datetime.combine(self.date, self.start)
        end = datetime.combine(self.date, self.end)
        return end - start


class AnnualLeave(TimeStampedModel, models.Model):
    """
    Model to keep track of staff employee annual leave

    This model is meant to be populated once a year
    Each staff member can only have one record per leave_type per year
    """
    YEAR_CHOICES = [
        (r, r) for r in range(2017, datetime.today().year + 10)
    ]

    year = models.PositiveIntegerField(
        _('Year'), choices=YEAR_CHOICES, default=2017, db_index=True)
    staff = models.ForeignKey(
        StaffProfile, verbose_name=_('Staff Member'), on_delete=models.CASCADE)
    leave_type = models.CharField(
        _('Type'), max_length=1, choices=Leave.TYPE_CHOICES, db_index=True)
    allowed_days = models.PositiveIntegerField(
        _('Allowed Leave days'), default=21, blank=True,
        help_text=_('Number of leave days allowed in a year.'))
    carried_over_days = models.PositiveIntegerField(
        _('Carried Over Leave days'), default=0, blank=True,
        help_text=_('Number of leave days carried over into this year.'))

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta options for AnnualLeave
        """
        verbose_name = _('Annual Leave')
        verbose_name_plural = _('Annual Leave')
        ordering = ['-year', 'leave_type', 'staff']
        unique_together = (('year', 'staff', 'leave_type'),)

    def __str__(self):
        # pylint: disable=no-member
        return _(
            f'{self.year}: {self.staff.get_name()} '
            f'{self.get_leave_type_display()}')

    def get_cumulative_leave_taken(self):
        """
        Get the cumulative leave taken

        Returns a timedelta
        """
        return get_taken_leave_days(
            staffprofile=self.staff,
            status=Leave.APPROVED,
            leave_type=self.leave_type,
            start_year=self.year,
            end_year=self.year
        )

    def get_available_leave_days(self, month: int = 12):
        """
        Get the remaining leave days
        """
        if month <= 0:
            month = 1
        elif month > 12:
            month = 12

        # the max allowed days per year
        allowed = self.allowed_days

        # the days `earned` per month
        per_month = Decimal(allowed / 12)

        # the days earned so far, given the month
        earned = Decimal(month) * per_month

        # the days taken
        taken = self.get_cumulative_leave_taken()

        # the starting balance
        starting_balance = self.carried_over_days

        return Decimal(earned + starting_balance - taken)


class FreeDay(models.Model):
    """Model definition for FreeDay."""
    name = models.CharField(_("Name"), max_length=255)
    date = models.DateField(_('Date'), unique=True)

    class Meta:
        """Meta definition for FreeDay."""
        ordering = ['-date']
        verbose_name = _('Free Day')
        verbose_name_plural = _('Free Days')

    def __str__(self):
        """Unicode representation of FreeDay."""
        return f"{self.date.year} - {self.name}"


def get_days(start: object, end: object):
    """
    Yield the days between two datetime objects
    """
    current_tz = timezone.get_current_timezone()
    local_start = current_tz.normalize(start)
    local_end = current_tz.normalize(end)
    span = local_end.date() - local_start.date()
    for i in range(span.days + 1):
        yield local_start.date() + timedelta(days=i)


def get_taken_leave_days(
        staffprofile: object,
        status: str,
        leave_type: str,
        start_year: int,
        end_year: int):
    """
    Calculate the number of leave days actually taken,
    taking into account weekends and weekend policy
    """
    count = Decimal(0)
    free_days = FreeDay.objects.filter(
        date__year__gte=start_year, date__year__lte=end_year
    ).values_list('date', flat=True)
    queryset = Leave.objects.filter(
        staff=staffprofile,
        status=status,
        leave_type=leave_type).filter(
            Q(start__year__gte=start_year) | Q(end__year__lte=end_year))
    for leave_obj in queryset:
        days = get_days(start=leave_obj.start, end=leave_obj.end)
        for day in days:
            if day.year >= start_year and day.year <= end_year and\
                    day not in free_days:
                day_value = settings.SSHR_DAY_LEAVE_VALUES[day.isoweekday()]
                count = count + Decimal(day_value)
    return count
